"""Settings for reformater."""

from __future__ import annotations

import json
import logging
import sys
from itertools import chain

# get pyproject.toml settings
from pathlib import Path
from typing import TYPE_CHECKING

import yaml
from click import echo
from colorama import Fore
from HtmlTagNames import html_tag_names
from HtmlVoidElements import html_void_elements
from pathspec import PathSpec
from pathspec.patterns.gitwildmatch import GitWildMatchPatternError

if sys.version_info >= (3, 11):
    try:
        import tomllib
    except ImportError:
        # Help users on older alphas
        if not TYPE_CHECKING:
            import tomli as tomllib
else:
    import tomli as tomllib

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping

    from typing_extensions import Any, TypeVar

    _TMappingStrAny = TypeVar("_TMappingStrAny", bound=Mapping[str, Any])


logger = logging.getLogger(__name__)


def find_project_root(src: Path) -> Path:
    """Attempt to get the project root."""
    for directory in (src, *src.resolve().parents):
        if (directory / ".git").exists():
            return directory

        if (directory / ".hg").is_dir():
            return directory

        if (directory / "pyproject.toml").is_file():
            return directory

        if (directory / ".djlintrc").is_file():
            return directory

    return directory


def load_gitignore(root: Path) -> PathSpec:
    """Search upstream for a .gitignore file."""
    gitignore = root / ".gitignore"
    if gitignore.is_file():
        with gitignore.open(encoding="utf-8") as this_file:
            git_lines = this_file.readlines()
    else:
        git_lines = []

    try:
        return PathSpec.from_lines("gitwildmatch", git_lines)

    except GitWildMatchPatternError as e:
        echo(f"Could not parse {gitignore}: {e}", err=True)
        raise


def find_pyproject(root: Path) -> Path | None:
    """Search upstream for a pyproject.toml file."""
    pyproject = root / "pyproject.toml"

    if pyproject.is_file():
        return pyproject

    return None


def find_djlintrc(root: Path) -> Path | None:
    """Search upstream for a pyproject.toml file."""
    djlintrc = root / ".djlintrc"

    if djlintrc.is_file():
        return djlintrc

    return None


def find_djlint_rules(root: Path) -> Path | None:
    """Search upstream for a pyproject.toml file."""
    rules = root / ".djlint_rules.yaml"

    if rules.is_file():
        return rules

    return None


def load_pyproject_config(filepath: Path) -> Any:
    """Load djlint config from pyproject.toml"""
    data = tomllib.loads(filepath.resolve().read_text(encoding="utf-8"))
    return data.get("tool", {}).get("djlint", {})


def load_djlintrc_config(filepath: Path) -> Any:
    """Load djlint config from .djlintrc"""
    return json.loads(filepath.resolve().read_text(encoding="utf-8"))


def load_project_settings(src: Path, config: str | None) -> Any:
    """Load djlint config from pyproject.toml."""
    djlint_content: dict[str, Any] = {}

    if config:
        try:
            path = Path(config)
            if path.name == "pyproject.toml":
                djlint_content.update(load_pyproject_config(path))
            else:
                djlint_content.update(load_djlintrc_config(path))

        except Exception as error:
            logger.error(
                "%sFailed to load config file %s. %s",
                Fore.RED,
                Path(config).resolve(),
                error,
            )

    pyproject_file = find_pyproject(src)

    if pyproject_file:
        content = load_pyproject_config(pyproject_file)
        if content:
            djlint_content.update(content)
            return content
        logger.info("No pyproject.toml found.")

    djlintrc_file = find_djlintrc(src)

    if djlintrc_file:
        try:
            djlint_content.update(load_djlintrc_config(djlintrc_file))

        except Exception as error:
            logger.error("%sFailed to load .djlintrc file. %s", Fore.RED, error)

    return djlint_content


def validate_rules(
    rules: Iterable[_TMappingStrAny],
) -> Iterator[_TMappingStrAny]:
    """Validate a list of linter rules. Returns valid rules."""
    for rule in rules:
        warning = False
        name = rule["rule"].get("name", "undefined")
        if "name" not in rule["rule"]:
            warning = True
            echo(Fore.RED + "Warning: A rule is missing a name! 😢")
        if (
            "patterns" not in rule["rule"]
            and "python_module" not in rule["rule"]
        ):
            warning = True
            echo(
                Fore.RED
                + f"Warning: Rule {name} is missing a pattern or a python_module! 😢"
            )
        if "message" not in rule["rule"]:
            warning = True
            echo(Fore.RED + f"Warning: Rule {name} is missing a message! 😢")

        if not warning:
            yield rule


def load_custom_rules(src: Path) -> Any:
    """Load djlint config from pyproject.toml."""
    djlint_rules_file = find_djlint_rules(src)

    if djlint_rules_file:
        return yaml.load(
            djlint_rules_file.read_text(encoding="utf-8"),
            Loader=yaml.SafeLoader,
        )

    return ()


def build_custom_blocks(custom_blocks: str | None) -> str | None:
    """Build regex string for custom template blocks."""
    if custom_blocks:
        open_tags = {x.strip() + r"\b" for x in custom_blocks.split(",")}
        close_tags = {f"end{x}" for x in open_tags}
        return "|" + "|".join(sorted(open_tags | close_tags))
    return None


def build_ignore_blocks(ignore_blocks: str | None) -> str | None:
    """Build regex string for template blocks to not format."""
    if ignore_blocks:
        open_tags = {x.strip() + r"\b" for x in ignore_blocks.split(",")}
        close_tags = {f"end{x}" for x in open_tags}
        return "|".join(sorted(open_tags | close_tags))
    return None


def build_custom_html(custom_html: str | None) -> str | None:
    """Build regex string for custom HTML blocks."""
    if custom_html:
        return "|" + "|".join(x.strip() for x in custom_html.split(","))
    return None


class Config:
    """Djlint Config."""

    def __init__(
        self,
        src: str,
        *,
        ignore: str = "",
        extension: str = "",
        indent: int | None = None,
        quiet: bool = False,
        profile: str | None = None,
        require_pragma: bool = False,
        reformat: bool = False,
        check: bool = False,
        lint: bool = False,
        use_gitignore: bool = False,
        warn: bool = False,
        preserve_leading_space: bool = False,
        preserve_blank_lines: bool = False,
        format_css: bool = False,
        format_js: bool = False,
        configuration: str | None = None,
        statistics: bool = False,
        include: str = "",
        ignore_case: bool = False,
        ignore_blocks: str = "",
        custom_blocks: str = "",
        blank_line_after_tag: str = "",
        blank_line_before_tag: str = "",
        line_break_after_multiline_tag: bool = False,
        custom_html: str = "",
        exclude: str = "",
        extend_exclude: str = "",
        linter_output_format: str = "",
        max_line_length: int | None = None,
        max_attribute_length: int | None = None,
        format_attribute_template_tags: bool = False,
        per_file_ignores: tuple[tuple[str, str], ...] = (),
        indent_css: int | None = None,
        indent_js: int | None = None,
        close_void_tags: bool = False,
        no_line_after_yaml: bool = False,
        no_function_formatting: bool = False,
        no_set_formatting: bool = False,
        max_blank_lines: int | None = None,
    ) -> None:
        self.reformat = reformat
        self.check = check
        self.lint = lint
        self.warn = warn

        if src == "-":
            self.project_root = find_project_root(Path.cwd())
        else:
            self.project_root = find_project_root(Path(src))

        djlint_settings = load_project_settings(
            self.project_root, configuration
        )

        self.gitignore = load_gitignore(self.project_root)
        # custom configuration options

        self.use_gitignore: bool = use_gitignore or djlint_settings.get(
            "use_gitignore", False
        )
        self.extension: str = str(
            extension or djlint_settings.get("extension", "html")
        )
        self.quiet: bool = quiet or djlint_settings.get("quiet", False)
        self.require_pragma: bool = (
            require_pragma
            or str(djlint_settings.get("require_pragma", "false")).lower()
            == "true"
        )

        self.custom_blocks: str = str(
            build_custom_blocks(
                custom_blocks or djlint_settings.get("custom_blocks")
            )
            or ""
        )

        self.custom_html: str = str(
            build_custom_html(custom_html or djlint_settings.get("custom_html"))
            or ""
        )

        self.format_attribute_template_tags: bool = (
            format_attribute_template_tags
            or djlint_settings.get("format_attribute_template_tags", False)
        )

        self.preserve_leading_space: bool = (
            preserve_leading_space
            or djlint_settings.get("preserve_leading_space", False)
        )
        self.ignore_blocks: str | None = build_ignore_blocks(
            ignore_blocks or djlint_settings.get("ignore_blocks", "")
        )

        self.preserve_blank_lines: bool = (
            preserve_blank_lines
            or djlint_settings.get("preserve_blank_lines", False)
        )

        self.format_js: bool = format_js or djlint_settings.get(
            "format_js", False
        )

        self.js_config = (
            {"indent_size": indent_js}
            if indent_js
            else djlint_settings.get("js")
        ) or {}

        self.css_config = (
            {"indent_size": indent_css}
            if indent_css
            else djlint_settings.get("css")
        ) or {}

        self.format_css: bool = format_css or djlint_settings.get(
            "format_css", False
        )

        self.ignore_case: bool = ignore_case or djlint_settings.get(
            "ignore_case", False
        )

        self.close_void_tags: bool = close_void_tags or djlint_settings.get(
            "close_void_tags", False
        )
        self.no_line_after_yaml: bool = (
            no_line_after_yaml
            or djlint_settings.get("no_line_after_yaml", False)
        )
        self.no_set_formatting: bool = no_set_formatting or djlint_settings.get(
            "no_set_formatting", False
        )
        self.no_function_formatting: bool = (
            no_function_formatting
            or djlint_settings.get("no_function_formatting", False)
        )

        # ignore is based on input and also profile
        self.ignore: str = str(ignore or djlint_settings.get("ignore", ""))
        self.include: str = str(include or djlint_settings.get("include", ""))

        self.files: list[str] | None = djlint_settings.get("files", None)
        self.stdin = False

        # codes to exclude
        profile_dict: dict[str, tuple[str, ...]] = {
            "html": ("D", "J", "T", "N", "M"),
            "django": ("J", "N", "M"),
            "jinja": ("D", "N", "M"),
            "nunjucks": ("D", "J", "M"),
            "handlebars": ("D", "J", "N"),
            "golang": ("D", "J", "N", "M"),
            "angular": ("D", "J", "H012", "H026", "H028"),
        }

        self.profile_code: tuple[str, ...] = profile_dict.get(
            str(profile or djlint_settings.get("profile", "html")).lower(), ()
        )
        self.profile: str = str(
            profile or djlint_settings.get("profile", "all")
        ).lower()

        self.linter_output_format: str = (
            linter_output_format
            or djlint_settings.get(
                "linter_output_format", "{code} {line} {message} {match}"
            )
        )

        # load linter rules
        rule_set = validate_rules(
            chain(
                yaml.safe_load(
                    (Path(__file__).parent / "rules.yaml").read_text(
                        encoding="utf-8"
                    )
                ),
                load_custom_rules(self.project_root),
            )
        )

        self.linter_rules = tuple(
            x
            for x in rule_set
            if x["rule"]["name"] not in self.ignore.split(",")
            and not any(
                x["rule"]["name"].startswith(code) for code in self.profile_code
            )
            and self.profile not in x["rule"].get("exclude", set())
            and (
                x["rule"].get("default", True)
                or x["rule"]["name"] in self.include.split(",")
            )
        )

        self.statistics = statistics

        # base options
        default_indent = 4
        if not indent:
            try:
                indent = int(djlint_settings.get("indent", default_indent))
            except ValueError:
                echo(
                    Fore.RED
                    + f"Error: Invalid pyproject.toml indent value {djlint_settings['indent']}"
                )
                indent = default_indent
        self.indent_size = indent
        self.indent: str = indent * " "

        try:
            self.max_blank_lines = int(
                djlint_settings.get("max_blank_lines", max_blank_lines or 0)
            )
        except ValueError:
            echo(
                Fore.RED
                + f"Error: Invalid pyproject.toml indent value {djlint_settings['max_blank_lines']}"
            )
            self.max_blank_lines = max_blank_lines or 0

        # From ruff and black
        default_exclude: str = r"""
            __pypackages__
            | _build
            | \.bzr
            | \.direnv
            | \.eggs
            | \.git
            | \.git-rewrite
            | \.hg
            | \.ipynb_checkpoints
            | \.mypy_cache
            | \.nox
            | \.pants\.d
            | \.pytest_cache
            | \.pytype
            | \.ruff_cache
            | \.svn
            | \.tox
            | \.venv
            | \.vscode
            | buck-out
            | build
            | dist
            | node_modules
            | venv
        """

        self.exclude: str = exclude or djlint_settings.get(
            "exclude", default_exclude
        )

        extend_exclude = extend_exclude or djlint_settings.get(
            "extend_exclude", ""
        )

        if extend_exclude:
            self.exclude += r" | " + r" | ".join(
                x.strip() for x in extend_exclude.split(",")
            )

        self.per_file_ignores = (
            (dict(per_file_ignores))
            if per_file_ignores
            else djlint_settings.get("per-file-ignores", {})
        )

        # add blank line after load tags
        self.blank_line_after_tag: str | None = (
            blank_line_after_tag
            or djlint_settings.get("blank_line_after_tag", None)
        )

        # add blank line before load tags
        self.blank_line_before_tag: str | None = (
            blank_line_before_tag
            or djlint_settings.get("blank_line_before_tag", None)
        )

        # add line break after multi-line tags
        self.line_break_after_multiline_tag: bool = (
            line_break_after_multiline_tag
            or djlint_settings.get("line_break_after_multiline_tag", False)
        )

        # contents of tags will not be formatted
        self.script_style_opening: str = r"""
           <style
         | <script
        """
        self.ignored_block_opening: str = r"""
              <style
            | {\*
            | <\?php
            | <script
            | <!--
            | [^\{]{\#(?!\s*djlint\:\s*(?:on|off))
            | ^{\#(?!\s*djlint\:\s*(?:on|off))
            | <pre
            | <textarea
            | {%[ ]*?blocktrans(?:late)?(?:(?!%}|\btrimmed\b).)*?%}
            | {\#\s*djlint\:\s*off\s*\#}
            | {%[ ]+?comment[ ]+?(?:(?!%}).)*?%}
            | {{!--\s*djlint\:off\s*--}}
            | {{-?\s*/\*\s*djlint\:off\s*\*/\s*-?}}
        """
        self.script_style_closing: str = r"""
              </style
            | </script
        """
        self.ignored_block_closing: str = r"""
              </style
            | \*}
            | \?>
            | </script
            |  -->
            | ^(?:(?!{\#).)*\#} # lines that have a #}, but not a {#
            | </pre
            | </textarea
            | {\#\s*djlint\:\s*on\s*\#}
            | (?<!djlint:off\s*?){%[ ]+?endcomment[ ]+?%}
            | {{!--\s*djlint\:on\s*--}}
            | {{-?\s*/\*\s*djlint\:on\s*\*/\s*-?}}
            | {%[ ]*?endblocktrans(?:late)?(?:(?!%}).)*?%}
        """

        # ignored block closing tags that
        # we can safely indent.
        self.safe_closing_tag: str = r"""
              </script
            | </style
            | {\#\s*djlint\:\s*on\s*\#}
            | {%[ ]+?endcomment[ ]+?%}
            | {{!--\s*djlint\:on\s*--}}
            | {{-?\s*/\*\s*djlint\:on\s*\*/\s*-?}}
        """

        # all html tags possible
        self.indent_html_tags: str = "|".join(html_tag_names) + self.custom_html

        self.indent_template_tags: str = (
            (rf"(?!{self.ignore_blocks})" if self.ignore_blocks else "")
            + r""" (?:if
                | ifchanged
                | for
                | asyncEach
                | asyncAll
                | block(?!trans|translate)
                | spaceless
                | compress
                | addto
                | language
                | with
                | assets
                | verbatim
                | autoescape
                | filter
                | each
                | macro
                | call
                | raw
                | blocktrans(?!late)
                | blocktranslate
                | thumbnail
                | set(?!(?:(?!%}).)*=)
            """
            + self.custom_blocks
            + r")\b"
        )

        self.template_indent: str = (
            r"""
            (?:\{\{\#|\{%-?)[ ]*?
                ("""
            + self.indent_template_tags
            + r"""
            ) | \{{-?[ ]*?form_start
            """
        )

        self.template_unindent: str = r"""
                (?:
                  (?:\{\{\/)
                | (?:\{%-?[ ]*?end(?!comment))
                | (?:\{{-?[ ]*?form_end)
              )
            """

        # these tags should be unindented and next line will be indented
        self.tag_unindent_line: str = r"""
              (?:\{%-?[ ]*?(?:elif|else|empty|plural))
            | (?:
                \{\{[ ]*?
                (
                    (?:else|\^)
                    [ ]*?\}\}
                )
              )
        """

        self.break_before = r"(?<!\n[ \t]*?)"

        # if lines are longer than x
        self.max_line_length = 120

        try:
            self.max_line_length = max_line_length or int(
                djlint_settings.get("max_line_length", self.max_line_length)
            )
        except ValueError:
            echo(
                Fore.RED
                + f"Error: Invalid pyproject.toml max_line_length value {djlint_settings['max_line_length']}"
            )

        self.max_attribute_length = 70

        try:
            self.max_attribute_length = max_attribute_length or int(
                djlint_settings.get(
                    "max_attribute_length", self.max_attribute_length
                )
            )
        except ValueError:
            echo(
                Fore.RED
                + f"Error: Invalid pyproject.toml max_attribute_length value {djlint_settings['max_attribute_length']}"
            )

        self.template_if_for_pattern = r"(?:{%-?\s?(?:if|for|asyncAll|asyncEach)[^}]*?%}(?:.*?{%\s?end(?:if|for|each|all)[^}]*?-?%})+?)"

        self.attribute_pattern: str = (
            rf"""
            (?:
                (
                    (?:\w|-|\.|\:|@|/(?!>))+ | required | checked
                )? # attribute name
                (?:  [ ]*?=[ ]*? # followed by "="
                    (
                        \"[^\"]*? # double quoted attribute
                        (?:
                            {self.template_if_for_pattern} # if or for loop
                           | {{{{.*?}}}} # template stuff
                           | {{%[^}}]*?%}}
                           | [^\"] # anything else
                        )*?
                        \" # closing quote
                      | '[^']*? # single quoted attribute
                        (?:
                            {self.template_if_for_pattern} # if or for loop
                           | {{{{.*?}}}} # template stuff
                           | {{%[^}}]*?%}}
                           | [^'] # anything else
                        )*?
                        \' # closing quote
                      | (?:\w|-)+ # or a non-quoted string value
                      | {{{{.*?}}}} # a non-quoted template var
                      | {{%[^}}]*?%}} # a non-quoted template tag
                      | {self.template_if_for_pattern} # a non-quoted if statement

                    )
                )? # attribute value
            )
            | ({self.template_if_for_pattern}
            """
            r"""
            | (?:\'|\") # allow random trailing quotes
            | {{.*?}}
            | {\#.*?\#}
            | {%.*?%})
        """
        )

        self.html_tag_regex = r"""
            (</?(?:!(?!--))?) # an opening bracket (< or </ or <!), but not a comment
            ([^\s>!\[]+\b) # a tag name
            ((?:\s*?(?:\"[^\"]*\"|'[^']*'|{{(?:(?!}}).)*}}|{%(?:(?!%}).)*%}|[^'\">{}/\s]|/(?!>)))+)? # any attributes
            \s*? # potentially some whitespace
            (/?>) # a closing bracket (/> or >)
        """

        self.attribute_style_pattern: str = (
            r"^(.*?)(style=)([\"|'])(([^\"']+?;)+?)\3"
        )

        self.ignored_attributes = frozenset({
            "href",
            "action",
            "data-url",
            "src",
            "url",
            "srcset",
            "data-src",
        })

        self.start_template_tags: str = (
            (rf"(?!{self.ignore_blocks})" if self.ignore_blocks else "")
            + r"""
              (?:if
            | for
            | asyncEach
            | asyncAll
            | block(?!trans)
            | spaceless
            | compress
            | load
            | assets
            | addto
            | language
            | with
            | assets
            | autoescape
            | filter
            | verbatim
            | each
            | macro
            | call
            | raw
            | blocktrans(?!late)
            | blocktranslate
            | thumbnail
            | set(?!(?:(?!%}).)*=)

            """
            + self.custom_blocks
            + r""")
        """
        )

        self.break_template_tags: str = (
            (rf"(?!{self.ignore_blocks})" if self.ignore_blocks else "")
            + r"""
              (?:if
            | endif
            | for
            | endfor
            | asyncEach
            | endeach
            | asyncAll
            | endall
            | block(?!trans)
            | endblock(?!trans)
            | else
            | plural
            | spaceless
            | endspaceless
            | compress
            | endcompress
            | load
            | include
            | assets
            | endassets
            | addto
            | language
            | with
            | endwith
            | autoescape
            | endautoescape
            | filter
            | endfilter
            | elif
            | resetcycle
            | verbatim
            | endverbatim
            | each
            | macro
            | endmacro
            | raw
            | endraw
            | call
            | endcall
            | image
            | blocktrans(?!late)
            | endblocktrans(?!late)
            | blocktranslate
            | endblocktranslate
            | set(?!(?:(?!%}).)*=)
            | endset
            | thumbnail
            | endthumbnail
            """
            + self.custom_blocks
            + r""")
        """
        )
        self.template_blocks: str = r"""
        {%((?!%}).)+%}|{{((?!}}).)+}}
        """

        self.ignored_linter_blocks: str = r"""
           {%-?[ ]*?raw\b(?:(?!%}).)*?-?%}.*?(?={%-?[ ]*?endraw[ ]*?-?%})
        """

        self.unformatted_blocks: str = r"""
            # html comment
            | <!--\s*djlint\:off\s*-->.(?:(?!<!--\s*djlint\:on\s*-->).)*
            # django/jinja/nunjucks
            | {\#\s*djlint\:\s*off\s*\#}(?:(?!{\#\s*djlint\:\s*on\s*\#}).)*
            | {%\s*comment\s*%\}\s*djlint\:off\s*\{%\s*endcomment\s*%\}(?:(?!{%\s*comment\s*%\}\s*djlint\:on\s*\{%\s*endcomment\s*%\}).)*
            # inline jinja comments
            | {\#(?!\s*djlint\:\s*(?:off|on)).*?\#}
            # handlebars
            | {{!--\s*djlint\:off\s*--}}(?:(?!{{!--\s*djlint\:on\s*--}}).)*
            # golang
            | {{-?\s*/\*\s*djlint\:off\s*\*/\s*-?}}(?:(?!{{-?\s*/\*\s*djlint\:on\s*\*/\s*-?}}).)*
            | ^---[\s\S]+?---
        """

        self.ignored_blocks: str = r"""
              <(pre|textarea).*?</(\1)>
            | <(script|style).*?(?=(\</(?:\3)>))
            # html comment
            | <!--\s*djlint\:off\s*-->.(?:(?!<!--\s*djlint\:on\s*-->).)*
            # django/jinja/nunjucks
            | {\#\s*djlint\:\s*off\s*\#}(?:(?!{\#\s*djlint\:\s*on\s*\#}).)*
            | {%\s*comment\s*%\}\s*djlint\:off\s*\{%\s*endcomment\s*%\}(?:(?!{%\s*comment\s*%\}\s*djlint\:on\s*\{%\s*endcomment\s*%\}).)*
            # inline jinja comments
            | {\#(?!\s*djlint\:\s*(?:off|on)).*?\#}
            # handlebars
            | {{!--\s*djlint\:off\s*--}}(?:(?!{{!--\s*djlint\:on\s*--}}).)*
            # golang
            | {{-?\s*/\*\s*djlint\:off\s*\*/\s*-?}}(?:(?!{{-?\s*/\*\s*djlint\:on\s*\*/\s*-?}}).)*
            # inline golang comments
            | {{-?\s*/\*(?!\s*djlint\:\s*(?:off|on)).*?\*/\s*-?}}
            | <!--.*?-->
            | <\?php.*?\?>
            | {%[ ]*?blocktranslate\b(?:(?!%}|\btrimmed\b).)*?%}.*?{%[ ]*?endblocktranslate[ ]*?%}
            | {%[ ]*?blocktrans\b(?:(?!%}|\btrimmed\b).)*?%}.*?{%[ ]*?endblocktrans[ ]*?%}
            | {%[ ]*?comment\b(?:(?!%}).)*?%}(?:(?!djlint:(?:off|on)).)*?(?={%[ ]*?endcomment[ ]*?%})
            | ^---[\s\S]+?---
        """
        self.script_style_inline: str = r"""
        <(script|style).*?(?=(\</(?:\1)>))
        """
        self.ignored_blocks_inline: str = r"""
              <(pre|textarea).*?</(\1)>
            | <(script|style).*?(?=(\</(?:\3)>))
            # html comment
            | <!--\s*djlint\:off\s*-->.*?(?=<!--\s*djlint\:on\s*-->)
            # django/jinja/nunjucks
            | {\#\s*djlint\:\s*off\s*\#}.*?(?={\#\s*djlint\:\s*on\s*\#})
            | {%\s*comment\s*%\}\s*djlint\:off\s*\{%\s*endcomment\s*%\}.*?(?={%\s*comment\s*%\}\s*djlint\:on\s*\{%\s*endcomment\s*%\})
            # inline jinja comments
            | {\#(?!\s*djlint\:\s*(?:off|on)).*?\#}
            # handlebars
            | {{!--\s*djlint\:off\s*--}}.*?(?={{!--\s*djlint\:on\s*--}})
            # golang
            | {{-?\s*/\*\s*djlint\:off\s*\*/\s*-?}}.*?(?={{-?\s*/\*\s*djlint\:on\s*\*/\s*-?}})
            # inline golang comments
            | {{-?\s*/\*(?!\s*djlint\:\s*(?:off|on)).*?\*/\s*-?}}
            | <!--.*?-->
            | <\?php.*?\?>
            | {%[ ]*?blocktranslate\b(?:(?!%}|\btrimmed\b).)*?%}.*?{%[ ]*?endblocktranslate[ ]*?%}
            | {%[ ]*?blocktrans\b(?:(?!%}|\btrimmed\b).)*?%}.*?{%[ ]*?endblocktrans[ ]*?%}
            | {%[ ]*?comment\b(?:(?!%}).)*?%}(?:(?!djlint:(?:off|on)).)*?(?={%[ ]*?endcomment[ ]*?%})
            | ^---[\s\S]+?---
        """

        self.ignored_rules = (
            # html comment
            r"<!--\s*djlint\:off(.+?)-->(?:(?!<!--\s*djlint\:on\s*-->).)*",
            # django/jinja/nunjucks
            r"{\#\s*djlint\:\s*off(.+?)\#}(?:(?!{\#\s*djlint\:\s*on\s*\#}).)*",
            r"{%\s*comment\s*%\}\s*djlint\:off(.*?)\{%\s*endcomment\s*%\}(?:(?!{%\s*comment\s*%\}\s*djlint\:on\s*\{%\s*endcomment\s*%\}).)*",
            # handlebars
            r"{{!--\s*djlint\:off(.*?)--}}(?:(?!{{!--\s*djlint\:on\s*--}}).)*",
            # golang
            r"{{-?\s*/\*\s*djlint\:off(.*?)\*/\s*-?}}(?:(?!{{-?\s*/\*\s*djlint\:on\s*\*/\s*-?}}).)*",
        )

        self.ignored_trans_blocks: str = r"""
              {%[ ]*?blocktranslate?\b(?:(?!%}|\btrimmed\b).)*?%}.*?{%[ ]*?endblocktranslate?[ ]*?%}
            | {%[ ]*?blocktrans\b(?:(?!%}|\btrimmed\b).)*?%}.*?{%[ ]*?endblocktrans[ ]*?%}
        """
        self.trans_trimmed_blocks: str = r"""
              {%[ ]*?blocktranslate\b(?:(?!%}).)*?\btrimmed\b(?:(?!%}).)*?%}.*?{%[ ]*?endblocktranslate[ ]*?%}
            | {%[ ]*?blocktrans\b(?:(?!%}).)*?\btrimmed\b(?:(?!%}).)*?%}.*?{%[ ]*?endblocktrans[ ]*?%}
        """
        self.ignored_trans_blocks_closing: str = r"""
         {%[ ]*?endblocktrans(?:late)?(?:(?!%}).)*?%}
        """

        self.ignored_inline_blocks: str = r"""
              <!--.*?-->
            | <script.*?\</script>
            | <style.*?\</style>
            | {\*.*?\*}
            | {\#(?!.*djlint:[ ]*?(?:off|on)\b).*\#}
            | <\?php.*?\?>
            | {%[ ]*?comment\b(?:(?!%}).)*?%}(?:(?!djlint:(?:off|on)).)*?{%[ ]*?endcomment[ ]*?%}
            | {%[ ]*?blocktrans(?:late)?\b(?:(?!%}|\btrimmed\b).)*?%}.*?{%[ ]*?endblocktrans(?:late)?[ ]*?%}
        """

        self.optional_single_line_html_tags: str = r"""
              button
            | a
            | h1
            | h2
            | h3
            | h4
            | h5
            | h6
            | td
            | th
            | strong
            | small
            | em
            | icon
            | span
            | title
            | link
            | path
            | label
            | div
            | li
            | script
            | style
            | head
            | body
            | p
            | select
            | article
            | option
            | legend
            | summary
            | dt
            | figcaption
            | tr
            | li
        """

        self.always_self_closing_html_tags: str = "|".join(html_void_elements)

        self.optional_single_line_template_tags: str = r"""
              if
            | for
            | block
            | with
            | asyncEach
            | asyncAll
        """

        self.break_html_tags: str = (
            r"""
              html
            | head
            | body
            | div
         #   | a # a gets no breaks #177
            | nav
            | ul
            | ol
            | dl
            | dd
            | dt
            | li
            | table
            | thead
            | tbody
            | tr
            | th
            | td
            | blockquote
            | select
            | form
            | option
            | optgroup
            | fieldset
            | legend
            | label
            | header
            | cache
            | main
            | section
            | aside
            | footer
            | figure
            | figcaption
            | video
         #   | span # span gets no breaks #171
            | p
            | g
            | svg
            | h\d
            | button
            | path
            | picture
            | script
            | style
            | details
            | summary
            | """
            + self.always_self_closing_html_tags
            + self.custom_html
            + """
        """
        )

        # the contents of these tag blocks will be indented, then unindented
        self.tag_indent: str = (
            self.template_indent
            + """
            | (?:<
                (?:
                    """
            + self.indent_html_tags
            + """
                )\\b
              )
        """
        )

        # either a template tag at the start of a line,
        # a html tag at the start of a line,
        # or an html tag as the end of a line.
        # Nothing in between!
        self.tag_unindent: str = (
            r"""
                ^
                """
            + self.template_unindent
            + """
            | (?:^</
                (?:
                    """
            + self.indent_html_tags
            + """
                )\\b
              )
            | (?:</
                (?:
                    """
            + self.indent_html_tags
            + """
                )>$
              )
        """
        )



"""Djlint reformat html files.

Much code is borrowed from https://github.com/rareyman/HTMLBeautify, many thanks!
"""

from __future__ import annotations

import difflib
from typing import TYPE_CHECKING

from .formatter.compress import compress_html
from .formatter.condense import clean_whitespace, condense_html
from .formatter.css import format_css
from .formatter.expand import expand_html
from .formatter.indent import indent_html
from .formatter.js import format_js

if TYPE_CHECKING:
    from pathlib import Path

    from .settings import Config


def formatter(config: Config, rawcode: str) -> str:
    """Format a html string."""
    if not rawcode:
        return rawcode

    # naturalize the line breaks
    compressed = compress_html("\n".join(rawcode.splitlines()), config)

    expanded = expand_html(compressed, config)

    condensed = clean_whitespace(expanded, config)

    indented_code = indent_html(condensed, config)

    beautified_code = condense_html(indented_code, config)

    if config.format_css:
        beautified_code = format_css(beautified_code, config)

    if config.format_js:
        beautified_code = format_js(beautified_code, config)

    # preserve original line endings
    line_ending = rawcode.find("\n")
    if line_ending > -1 and rawcode[max(line_ending - 1, 0)] == "\r":
        # convert \r?\n to \r\n
        beautified_code = beautified_code.replace("\r", "").replace(
            "\n", "\r\n"
        )

    return beautified_code


def reformat_file(
    config: Config, this_file: Path
) -> dict[str, tuple[str, ...]]:
    """Reformat html file."""
    with this_file.open(encoding="utf-8", newline="") as f:
        rawcode = f.read()

    beautified_code = formatter(config, rawcode)

    if (
        config.check is not True and beautified_code != rawcode
    ) or config.stdin:
        with this_file.open("w", encoding="utf-8", newline="") as f:
            f.write(beautified_code)

    return {
        str(this_file): tuple(
            difflib.unified_diff(
                rawcode.splitlines(), beautified_code.splitlines()
            )
        )
    }



"""Compress html.

1. flatten attributes
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import regex as re
from HtmlTagNames import html_tag_names
from HtmlVoidElements import html_void_elements

from ..helpers import child_of_unformatted_block

if TYPE_CHECKING:
    from ..settings import Config


def compress_html(html: str, config: Config) -> str:
    """Compress html."""

    def _fix_case(tag: str) -> str:
        if not config.ignore_case and tag.lower() in html_tag_names:
            return tag.lower()

        if not config.ignore_case and tag.lower() == "doctype":
            return "DOCTYPE"
        return tag

    def _clean_tag(match: re.Match[str]) -> str:
        """Flatten multiline attributes back to one line.

        Skip when attribute is ignored.
        Attribute name can be in group one or group 2.
        for now, skipping if they are anywhere

        tags starting ignored blocks can have their attributes formatted,
        for example <textarea class="..." id="..."> can be formatted.
        """
        if child_of_unformatted_block(config, html, match):
            return match.group()

        open_bracket = match.group(1)
        tag = _fix_case(match.group(2))

        attributes = (
            (
                " "
                + " ".join(
                    x.strip() for x in match.group(3).strip().splitlines()
                )
            )
            if match.group(3)
            else ""
        )
        if tag.lower() in html_void_elements and config.close_void_tags:
            close_bracket = " />"
        else:
            close_bracket = (
                match.group(4)
                if "/" not in match.group(4)
                else f" {match.group(4)}"
            )

        return f"{open_bracket}{tag}{attributes}{close_bracket}"

    return re.sub(
        config.html_tag_regex,
        _clean_tag,
        html,
        flags=re.MULTILINE | re.VERBOSE | re.IGNORECASE,
    )



"""djLint · lint and reformat HTML templates."""

from __future__ import annotations

import os
import sys
import tempfile
from concurrent.futures import (
    ProcessPoolExecutor,
    ThreadPoolExecutor,
    as_completed,
)
from pathlib import Path
from typing import TYPE_CHECKING

import click
from click import echo
from colorama import Fore, Style, colorama_text
from tqdm import tqdm

from .lint import lint_file
from .output import print_output
from .reformat import reformat_file
from .settings import Config
from .src import get_src

if TYPE_CHECKING:
    from .output import ProcessResult


@click.command(context_settings={"help_option_names": ["-h", "--help"]})  # type: ignore[operator]
@click.argument(
    "src",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
        allow_dash=True,
    ),
    nargs=-1,
    required=True,
    metavar="SRC ...",
)
@click.version_option(package_name="djlint")
@click.option(
    "-e",
    "--extension",
    type=str,
    default="",
    help="File extension to check [default: html]",
    show_default=False,
)
@click.option(
    "-i",
    "--ignore",
    type=str,
    default="",
    help='Codes to ignore. ex: "H014,H017"',
    show_default=False,
)
@click.option("--reformat", is_flag=True, help="Reformat the file(s).")
@click.option("--check", is_flag=True, help="Check formatting on the file(s).")
@click.option(
    "--indent",
    type=int,
    help="Indent spacing. [default: 4]",
    show_default=False,
)
@click.option(
    "--quiet", is_flag=True, help="Do not print diff when reformatting."
)
@click.option(
    "--profile",
    type=str,
    help="Enable defaults by template language. ops: django, jinja, nunjucks, handlebars, golang, angular, html [default: html]",
)
@click.option(
    "--require-pragma",
    is_flag=True,
    help="Only format or lint files that starts with a comment with the text 'djlint:on'",
)
@click.option(
    "--lint", is_flag=True, help="Lint for common issues. [default option]"
)
@click.option(
    "--use-gitignore",
    is_flag=True,
    help="Use .gitignore file to extend excludes.",
)
@click.option("--warn", is_flag=True, help="Return errors as warnings.")
@click.option(
    "--preserve-leading-space",
    is_flag=True,
    help="Attempt to preserve leading space on text.",
)
@click.option(
    "--preserve-blank-lines",
    is_flag=True,
    help="Attempt to preserve blank lines.",
)
@click.option(
    "--format-css", is_flag=True, help="Also format contents of <style> tags."
)
@click.option(
    "--format-js", is_flag=True, help="Also format contents of <script> tags."
)
@click.option(
    "--configuration",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
        allow_dash=True,
    ),
    required=False,
    help="Path to global configuration file in .djlintrc format",
)
@click.option(
    "--statistics",
    is_flag=True,
    help="Count the number of occurrences of each error/warning code.",
)
@click.option(
    "--include",
    type=str,
    default="",
    help='Codes to include. ex: "H014,H017"',
    show_default=False,
)
@click.option(
    "--ignore-case", is_flag=True, help="Do not fix case on known html tags."
)
@click.option(
    "--ignore-blocks",
    type=str,
    default="",
    help="Comma list of template blocks to not indent.",
)
@click.option(
    "--blank-line-after-tag",
    type=str,
    default="",
    help="Add an additional blank line after {% <tag> ... %} tag groups.",
)
@click.option(
    "--blank-line-before-tag",
    type=str,
    default="",
    help="Add an additional blank line before {% <tag> ... %} tag groups.",
)
@click.option(
    "--line-break-after-multiline-tag",
    is_flag=True,
    help="Do not condense the content of multi-line tags into the line of the last attribute.",
)
@click.option(
    "--custom-blocks",
    type=str,
    default="",
    help="Indent custom template blocks. For example {% toc %}...{% endtoc %}",
)
@click.option(
    "--custom-html",
    type=str,
    default="",
    help="Indent custom HTML tags. For example <mjml>",
)
@click.option(
    "--exclude",
    type=str,
    default="",
    help="Override the default exclude paths.",
)
@click.option(
    "--extend-exclude",
    type=str,
    default="",
    help="Add additional paths to the default exclude.",
)
@click.option(
    "--linter-output-format",
    type=str,
    default="",
    help="Customize order of linter output message.",
)
@click.option(
    "--max-line-length",
    type=int,
    help="Max line length. [default: 120]",
    show_default=False,
)
@click.option(
    "--max-attribute-length",
    type=int,
    help="Max attribute length. [default: 70]",
    show_default=False,
)
@click.option(
    "--format-attribute-template-tags",
    is_flag=True,
    help="Attempt to format template syntax inside of tag attributes.",
)
@click.option(
    "--per-file-ignores",
    type=(str, str),
    multiple=True,
    help="Ignore linter rules on a per-file basis.",
)
@click.option(
    "--indent-css", type=int, help="Set CSS indent level.", show_default=False
)
@click.option(
    "--indent-js", type=int, help="Set JS indent level.", show_default=False
)
@click.option(
    "--close-void-tags",
    is_flag=True,
    help="Add closing mark on known void tags. Ex: <img> becomse <img />",
)
@click.option(
    "--no-line-after-yaml",
    is_flag=True,
    help="Do not add a blank line after yaml front matter.",
)
@click.option(
    "--no-function-formatting",
    is_flag=True,
    help="Do not attempt to format function contents.",
)
@click.option(
    "--no-set-formatting",
    is_flag=True,
    help="Do not attempt to format set contents.",
)
@click.option(
    "--max-blank-lines",
    type=int,
    help="Consolidate blank lines down to x lines. [default: 0]",
    show_default=False,
)
@colorama_text(autoreset=True)
def main(
    *,
    src: tuple[str, ...],
    extension: str,
    ignore: str,
    reformat: bool,
    indent: int | None,
    check: bool,
    quiet: bool,
    profile: str | None,
    require_pragma: bool,
    lint: bool,
    use_gitignore: bool,
    warn: bool,
    preserve_leading_space: bool,
    preserve_blank_lines: bool,
    format_css: bool,
    format_js: bool,
    configuration: str | None,
    statistics: bool,
    include: str,
    ignore_case: bool,
    ignore_blocks: str,
    blank_line_after_tag: str,
    blank_line_before_tag: str,
    line_break_after_multiline_tag: bool,
    custom_blocks: str,
    custom_html: str,
    exclude: str,
    extend_exclude: str,
    linter_output_format: str,
    max_line_length: int | None,
    max_attribute_length: int | None,
    format_attribute_template_tags: bool,
    per_file_ignores: tuple[tuple[str, str], ...],
    indent_css: int | None,
    indent_js: int | None,
    close_void_tags: bool,
    no_line_after_yaml: bool,
    no_function_formatting: bool,
    no_set_formatting: bool,
    max_blank_lines: int | None,
) -> None:
    """djLint · HTML template linter and formatter."""
    config = Config(
        src[0],
        extension=extension,
        ignore=ignore,
        indent=indent,
        quiet=quiet,
        profile=profile,
        require_pragma=require_pragma,
        lint=lint or not (reformat or check),
        reformat=reformat,
        check=check,
        use_gitignore=use_gitignore,
        warn=warn,
        preserve_leading_space=preserve_leading_space,
        preserve_blank_lines=preserve_blank_lines,
        format_css=format_css,
        format_js=format_js,
        configuration=configuration,
        statistics=statistics,
        include=include,
        ignore_case=ignore_case,
        ignore_blocks=ignore_blocks,
        blank_line_after_tag=blank_line_after_tag,
        blank_line_before_tag=blank_line_before_tag,
        line_break_after_multiline_tag=line_break_after_multiline_tag,
        custom_blocks=custom_blocks,
        custom_html=custom_html,
        exclude=exclude,
        extend_exclude=extend_exclude,
        linter_output_format=linter_output_format,
        max_line_length=max_line_length,
        max_attribute_length=max_attribute_length,
        format_attribute_template_tags=format_attribute_template_tags,
        per_file_ignores=per_file_ignores,
        indent_css=indent_css,
        indent_js=indent_js,
        close_void_tags=close_void_tags,
        no_line_after_yaml=no_line_after_yaml,
        no_function_formatting=no_function_formatting,
        no_set_formatting=no_set_formatting,
        max_blank_lines=max_blank_lines,
    )

    temp_file = None

    try:
        if "-" in src:
            if config.files:
                file_list = get_src((Path(x) for x in config.files), config)

            else:
                config.stdin = True
                stdin_stream = click.get_text_stream("stdin", encoding="utf-8")
                stdin_text = stdin_stream.read()
                temp_file = tempfile.NamedTemporaryFile(delete=False)  # noqa: SIM115
                temp_file.write(stdin_text.encode("utf-8"))
                temp_file.seek(0)

                # cannot use gitignore for stdin paths.
                config.use_gitignore = False

                file_list = get_src([Path(temp_file.name)], config)

        else:
            file_list = get_src((Path(x) for x in src), config)

        if not file_list:
            return

        message = ""

        if config.check:
            message = "Checking"
        elif config.reformat:
            message = "Reformatting"

        if config.lint:
            if message:
                message += " and "
            message += "Linting"

        bar_message = "{}{}{} {}{{n_fmt}}/{{total_fmt}}{} {}files{} {{bar}} {}{{elapsed}}{}".format(
            Fore.BLUE + Style.BRIGHT,
            message,
            Style.RESET_ALL,
            Fore.RED + Style.BRIGHT,
            Style.RESET_ALL,
            Fore.BLUE + Style.BRIGHT,
            Style.RESET_ALL,
            Fore.GREEN + Style.BRIGHT,
            Style.RESET_ALL + "    ",
        )
        if not config.stdin and not config.quiet:
            echo()

        progress_char = " »" if sys.platform == "win32" else "┈━"
        worker_count = min(os.cpu_count() or 1, len(file_list), 4)
        executor_cls = (
            ProcessPoolExecutor if worker_count > 1 else ThreadPoolExecutor
        )

        with executor_cls(max_workers=worker_count) as exe:
            futures = {
                exe.submit(process, config, this_file): this_file
                for this_file in file_list
            }

            if temp_file is None:
                file_errors = []
                elapsed = "00:00"
                with tqdm(
                    total=len(file_list),
                    bar_format=bar_message,
                    colour="BLUE",
                    ascii=progress_char,
                    leave=False,
                ) as pbar:
                    for future in as_completed(futures):
                        file_errors.append(future.result())
                        pbar.update()
                        elapsed = pbar.format_interval(
                            pbar.format_dict["elapsed"]
                        )

                finished_bar_message = f"{Fore.BLUE + Style.BRIGHT}{message}{Style.RESET_ALL} {Fore.GREEN + Style.BRIGHT}{{n_fmt}}/{{total_fmt}}{Style.RESET_ALL} {Fore.BLUE + Style.BRIGHT}files{Style.RESET_ALL} {{bar}} {Fore.GREEN + Style.BRIGHT}{elapsed}{Style.RESET_ALL}    "

                with tqdm(
                    total=len(file_list),
                    initial=len(file_list),
                    bar_format=finished_bar_message,
                    colour="GREEN",
                    ascii=progress_char,
                    leave=True,
                ):
                    pass
            else:
                file_errors = [
                    future.result() for future in as_completed(futures)
                ]

        if temp_file and (config.reformat or config.check):
            # if using stdin, only give back formatted code.
            echo(
                Path(temp_file.name)
                .read_text(encoding="utf-8")
                .rstrip()
                .encode("utf-8")
            )
    finally:
        if temp_file:
            try:
                temp_file.close()
            finally:
                Path(temp_file.name).unlink(missing_ok=True)

    if print_output(config, file_errors, len(file_list)) and not config.warn:
        sys.exit(1)


def process(config: Config, this_file: Path) -> ProcessResult:
    """Run linter or formatter."""
    output: ProcessResult = {}
    if config.reformat or config.check:
        output["format_message"] = reformat_file(config, this_file)

    if config.lint:
        output["lint_message"] = lint_file(config, this_file)

    return output



"""djLint function to call jsbeautifier."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

import jsbeautifier
import regex as re
from jsbeautifier.javascript.options import BeautifierOptions

from ..helpers import child_of_unformatted_block

if TYPE_CHECKING:
    from ..settings import Config


def format_js(html: str, config: Config) -> str:
    """Format javascript inside <script> tags."""

    def launch_formatter(
        config: Config, html: str, match: re.Match[str]
    ) -> str:
        """Add break after if not in ignored block."""
        if child_of_unformatted_block(config, html, match):
            return match.group()

        if not match.group(3).strip():
            return match.group()

        indent = len(match.group(1)) * " "

        # because of the param options for js-beautifier we cannot pass
        # in a fixed space leading.
        # so, call formatter twice, once with a fake indent.
        # check which lines changed (these are the formattable lines)
        # and add the leading space to them.

        config.js_config["indent_level"] = 1
        opts = BeautifierOptions(config.js_config)
        beautified_lines = jsbeautifier.beautify(
            match.group(3), opts
        ).splitlines()

        config.js_config["indent_level"] = 2
        opts = BeautifierOptions(config.js_config)
        beautified_lines_test = jsbeautifier.beautify(
            match.group(3), opts
        ).splitlines()

        beautified = ""
        for line, test in zip(beautified_lines, beautified_lines_test):
            beautified += "\n"
            if line == test:
                beautified += line
                continue
            beautified += indent + line

        return match.group(1) + match.group(2) + beautified + "\n" + indent

    func = partial(launch_formatter, config, html)

    return re.sub(
        r"([ ]*?)(<(?:script)\b(?:\"[^\"]*\"|'[^']*'|{[^}]*}|[^'\">{}])*>)(.*?)(?=</script>)",
        func,
        html,
        flags=re.IGNORECASE | re.DOTALL,
    )



"""Collection of shared djLint functions."""
# ruff: noqa: ERA001

from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

import regex as re

if TYPE_CHECKING:
    from .settings import Config


def is_ignored_block_opening(config: Config, item: str) -> bool:
    """Find ignored group opening.

    A valid ignored group opening tag will not be part of a
    single line block.
    """
    last_index = 0
    inline = tuple(
        re.finditer(
            config.ignored_blocks_inline,
            item,
            flags=re.IGNORECASE | re.VERBOSE | re.MULTILINE | re.DOTALL,
        )
    )

    if inline:
        last_index = (
            inline[-1].end()
        )  # get the last index. The ignored opening should start after this.

    return bool(
        re.search(
            config.ignored_block_opening,
            item[last_index:],
            flags=re.IGNORECASE | re.VERBOSE,
        )
    )


def is_script_style_block_opening(config: Config, item: str) -> bool:
    """Find ignored group opening.

    A valid ignored group opening tag will not be part of a
    single line block.
    """
    last_index = 0
    inline = tuple(
        re.finditer(
            config.script_style_inline,
            item,
            flags=re.IGNORECASE | re.VERBOSE | re.MULTILINE | re.DOTALL,
        )
    )

    if inline:
        last_index = (
            inline[-1].end()
        )  # get the last index. The ignored opening should start after this.

    return bool(
        re.search(
            config.script_style_opening,
            item[last_index:],
            flags=re.IGNORECASE | re.VERBOSE,
        )
    )


def inside_protected_trans_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Find ignored group closing.

    A valid ignored group closing tag will not be part of a
    single line block.

    True = non indentable > inside ignored trans block
    False = indentable > either inside a trans trimmed block, or somewhere else, but not a trans non trimmed :)
    """
    last_index = 0
    close_block = re.search(
        config.ignored_trans_blocks_closing,
        match.group(),
        flags=re.IGNORECASE | re.VERBOSE,
    )

    if not close_block:
        return False

    non_trimmed = tuple(
        re.finditer(
            config.ignored_trans_blocks,
            html,
            flags=re.IGNORECASE | re.VERBOSE | re.DOTALL,
        )
    )

    trimmed = tuple(
        re.finditer(
            config.trans_trimmed_blocks,
            html,
            flags=re.IGNORECASE | re.VERBOSE | re.DOTALL,
        )
    )

    # who is max?
    if non_trimmed and (
        not trimmed or non_trimmed[-1].end() > trimmed[-1].end()
    ):
        # non trimmed!
        # check that this is not an inline block.
        non_trimmed_inline = any(
            re.finditer(
                config.ignored_trans_blocks,
                match.group(),
                flags=re.IGNORECASE | re.VERBOSE | re.DOTALL,
            )
        )

        if non_trimmed_inline:
            last_index = non_trimmed[
                -1
            ].end()  # get the last index. The ignored opening should start after this.

            return bool(
                re.search(
                    config.ignored_trans_blocks_closing,
                    html[last_index:],
                    flags=re.IGNORECASE | re.VERBOSE,
                )
            )

        return close_block.end(0) <= non_trimmed[-1].end()

    if trimmed:
        # inside a trimmed block, we can return true to continue as if
        # this is a indentable block
        return close_block.end(0) > trimmed[-1].end()
    return False

    # print(close_block)
    # if non_trimmed:
    #     last_index = (
    #         non_trimmed[-1].end()
    #     )  # get the last index. The ignored opening should start after this.

    # return re.search(
    #     config.ignored_trans_blocks_closing,
    #     html[last_index:],
    #     flags=re.IGNORECASE | re.VERBOSE,
    # )


def is_ignored_block_closing(config: Config, item: str) -> bool:
    """Find ignored group closing.

    A valid ignored group closing tag will not be part of a
    single line block.
    """
    last_index = 0
    inline = tuple(
        re.finditer(
            config.ignored_inline_blocks, item, flags=re.IGNORECASE | re.VERBOSE
        )
    )

    if inline:
        last_index = (
            inline[-1].end()
        )  # get the last index. The ignored opening should start after this.

    return bool(
        re.search(
            config.ignored_block_closing,
            item[last_index:],
            flags=re.IGNORECASE | re.VERBOSE,
        )
    )


def is_script_style_block_closing(config: Config, item: str) -> bool:
    """Find ignored group closing.

    A valid ignored group closing tag will not be part of a
    single line block.
    """
    last_index = 0
    inline = tuple(
        re.finditer(
            config.script_style_inline, item, flags=re.IGNORECASE | re.VERBOSE
        )
    )

    if inline:
        last_index = (
            inline[-1].end()
        )  # get the last index. The ignored opening should start after this.

    return bool(
        re.search(
            config.script_style_closing,
            item[last_index:],
            flags=re.IGNORECASE | re.VERBOSE,
        )
    )


def is_safe_closing_tag(config: Config, item: str) -> bool:
    """Find ignored group opening.

    A valid ignored group opening tag will not be part of a
    single line block.
    """
    last_index = 0
    inline = tuple(
        re.finditer(
            config.ignored_inline_blocks + r" | " + config.ignored_blocks,
            item,
            flags=re.IGNORECASE | re.VERBOSE | re.MULTILINE | re.DOTALL,
        )
    )

    if inline:
        last_index = (
            inline[-1].end()
        )  # get the last index. The ignored opening should start after this.

    return bool(
        re.search(
            config.safe_closing_tag,
            item[last_index:],
            flags=re.IGNORECASE | re.VERBOSE,
        )
    )


def inside_template_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Check if a re.Match is inside of a template block."""
    return any(
        ignored_match.start(0) <= match.start()
        and match.end(0) <= ignored_match.end()
        for ignored_match in re.finditer(
            config.template_blocks,
            html,
            flags=re.DOTALL | re.IGNORECASE | re.VERBOSE | re.MULTILINE,
        )
    )


def inside_ignored_linter_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Check if a re.Match is inside of a ignored linter block."""
    return any(
        ignored_match.start(0) <= match.start()
        and match.end(0) <= ignored_match.end()
        for ignored_match in re.finditer(
            config.ignored_linter_blocks,
            html,
            flags=re.DOTALL | re.IGNORECASE | re.VERBOSE | re.MULTILINE,
        )
    )


def inside_ignored_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Do not add whitespace if the tag is in a non indent block."""
    return any(
        ignored_match.start(0) <= match.start()
        and match.end(0) <= ignored_match.end()
        for ignored_match in itertools.chain(
            re.finditer(
                config.ignored_blocks,
                html,
                flags=re.DOTALL | re.IGNORECASE | re.VERBOSE | re.MULTILINE,
            ),
            re.finditer(
                config.ignored_inline_blocks,
                html,
                flags=re.IGNORECASE | re.VERBOSE,
            ),
        )
    )


def child_of_unformatted_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Do not add whitespace if the tag is in a non indent block."""
    return any(
        ignored_match.start(0) < match.start()
        and match.end(0) <= ignored_match.end()
        for ignored_match in re.finditer(
            config.unformatted_blocks,
            html,
            flags=re.DOTALL | re.IGNORECASE | re.VERBOSE | re.MULTILINE,
        )
    )


def child_of_ignored_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Do not add whitespace if the tag is in a non indent block."""
    return any(
        ignored_match.start(0) < match.start()
        and match.end(0) <= ignored_match.end()
        for ignored_match in itertools.chain(
            re.finditer(
                config.ignored_blocks,
                html,
                flags=re.DOTALL | re.IGNORECASE | re.VERBOSE | re.MULTILINE,
            ),
            re.finditer(
                config.ignored_inline_blocks,
                html,
                flags=re.IGNORECASE | re.VERBOSE,
            ),
        )
    )


def overlaps_ignored_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Do not add whitespace if the tag is in a non indent block."""
    return any(
        # don't require the match to be fully inside the ignored block.
        # poorly build html will probably span ignored blocks and should be ignored.
        (
            ignored_match.start(0) <= match.start()
            and match.start() <= ignored_match.end()
        )
        or (
            ignored_match.start(0) <= match.end()
            and match.end() <= ignored_match.end()
        )
        for ignored_match in itertools.chain(
            re.finditer(
                config.ignored_blocks,
                html,
                flags=re.DOTALL
                | re.IGNORECASE
                | re.VERBOSE
                | re.MULTILINE
                | re.DOTALL,
            ),
            re.finditer(
                config.ignored_inline_blocks,
                html,
                flags=re.IGNORECASE | re.VERBOSE,
            ),
        )
    )


def inside_ignored_rule(
    config: Config, html: str, match: re.Match[str], rule: str
) -> bool:
    """Check if match is inside an ignored pattern."""
    for rule_regex in config.ignored_rules:
        for ignored_match in re.finditer(
            rule_regex, html, flags=re.DOTALL | re.IGNORECASE | re.VERBOSE
        ):
            if (
                rule in re.split(r"\s|,", ignored_match.group(1).strip())
                and (
                    ignored_match.start(0) <= match.start()
                    and match.start() <= ignored_match.end()
                )
            ) or (
                not ignored_match.group(1).strip()
                and (
                    ignored_match.start(0) <= match.end()
                    and match.end() <= ignored_match.end()
                )
            ):
                return True
    return False



"""Build djLint console output."""

from __future__ import annotations

import math
import shutil
import sys
from collections import Counter
from pathlib import Path
from typing import TYPE_CHECKING

import regex as re
from click import echo
from colorama import Fore, Style

if TYPE_CHECKING:
    from collections.abc import Collection, Iterable, Mapping, Sequence

    from typing_extensions import TypedDict

    from .lint import LintError
    from .settings import Config

    class ProcessResult(TypedDict, total=False):
        format_message: dict[str, tuple[str, ...]]
        lint_message: dict[str, list[LintError]]


try:
    sys.stdout.reconfigure(encoding="utf-8")  # type: ignore[union-attr]
except Exception:
    pass


def _count_digits(num: int, /) -> int:
    """Faster alternative to len(str(num))"""
    if num == 0:
        return 1
    return math.floor(math.log10(abs(num))) + 1


def print_output(
    config: Config, file_errors: Iterable[ProcessResult], file_count: int
) -> int:
    """Print results to console."""
    file_quantity = build_quantity(file_count)
    # format errors
    reformat_success_message = ""
    lint_success_message = ""
    lint_error_count = 0
    format_error_count = 0
    print_blanks = not config.stdin and not config.quiet

    if print_blanks:
        echo()

    for error in sorted(
        file_errors,
        key=lambda x: next(iter(next(iter(x.values())))),  # type: ignore[call-overload]
    ):
        if error.get("format_message") and not config.stdin:
            # reformat message
            format_error_count += build_check_output(
                error["format_message"], config
            )

        if error.get("lint_message"):
            # lint message
            lint_error_count += build_output(error["lint_message"], config)

    if config.statistics and config.lint:
        build_stats_output([x.get("lint_message") for x in file_errors], config)

    tense_message = (
        build_quantity(format_error_count) + " would be"
        if config.check
        else build_quantity_tense(format_error_count)
    )
    reformat_success_message = f"{tense_message} updated."

    error_case = "error" if lint_error_count == 1 else "errors"
    lint_success_message += (
        f"Linted {file_quantity}, found {lint_error_count} {error_case}."
    )

    if print_blanks:
        echo()

    if (
        not config.quiet
        and not config.stdin
        and (config.reformat or config.check)
    ):
        reformat_success_color = (
            Fore.RED + Style.BRIGHT if (format_error_count) > 0 else Fore.BLUE
        )
        echo(
            f"{reformat_success_color}{reformat_success_message}{Style.RESET_ALL}"
        )

    if config.lint and not config.quiet:
        lint_success_color = (
            Fore.RED + Style.BRIGHT if (lint_error_count) > 0 else Fore.BLUE
        )
        echo(f"{lint_success_color}{lint_success_message}{Style.RESET_ALL}")

    if print_blanks:
        echo()

    return lint_error_count + format_error_count


def build_relative_path(url: str, project_root: Path) -> str:
    """Get path relative to project."""
    url_path = Path(url)
    if project_root != url_path and project_root in url_path.parents:
        return str(url_path.relative_to(project_root.resolve()))

    return url


def build_output(
    error: Mapping[str, Iterable[LintError]], config: Config
) -> int:
    """Build output for file errors."""
    errors = sorted(
        next(iter(error.values())),
        key=lambda x: tuple(int(i) for i in x["line"].split(":")),
    )

    width, _ = shutil.get_terminal_size()

    if not errors:
        return 0

    filename = build_relative_path(
        next(iter(error.keys())), config.project_root.resolve()
    )

    if "{filename}" not in config.linter_output_format and not config.stdin:
        echo(
            f"{Fore.GREEN}{Style.BRIGHT}\n{filename}\n{Style.DIM}"
            + "".join("─" for _ in range(1, width))
            + Style.RESET_ALL
        )

    for message_dict in errors:
        line = Fore.BLUE + message_dict["line"] + Style.RESET_ALL
        code = (
            (Fore.RED if message_dict["code"][:1] == "E" else Fore.YELLOW)
            + message_dict["code"]
            + Style.RESET_ALL
        )
        message = message_dict["message"]
        match = (
            Fore.BLUE
            + re.sub(r"\s{2,}|\n", " ", message_dict["match"])
            + Style.RESET_ALL
        )

        echo(
            config.linter_output_format.format(
                filename=filename,
                line=line,
                code=code,
                message=message,
                match=match,
            ),
            err=False,
        )

    return len(errors)


def build_check_output(
    errors: Mapping[str, Sequence[str]], config: Config
) -> int:
    """Build output for reformat check."""
    if not errors:
        return 0

    color = {"-": Fore.YELLOW, "+": Fore.GREEN, "@": Style.BRIGHT + Fore.BLUE}
    width, _ = shutil.get_terminal_size()

    if not config.quiet and bool(next(iter(errors.values()))):
        echo(
            Fore.GREEN
            + Style.BRIGHT
            + "\n"
            + build_relative_path(
                next(iter(errors.keys())), config.project_root.resolve()
            )
            + "\n"
            + Style.DIM
            + "".join("─" for _ in range(1, width))
            + Style.RESET_ALL
        )

        for diff in next(iter(errors.values()))[2:]:
            echo(
                f"{color.get(diff[:1], Style.RESET_ALL)}{diff}{Style.RESET_ALL}",
                err=False,
            )

    return sum(1 for v in errors.values() if v)


def build_quantity(size: int) -> str:
    """Count files in a list."""
    return str(size) + " file" + ("s" if size > 1 or size == 0 else "")


def build_quantity_tense(size: int) -> str:
    """Count files in a list."""
    return (
        str(size)
        + " file"
        + ("s" if size > 1 or size == 0 else "")
        + " "
        + ("were" if size > 1 or size == 0 else "was")
    )


def build_stats_output(
    errors: Collection[Mapping[str, Iterable[LintError]] | None], config: Config
) -> int:
    """Build output for linter statistics."""
    if not errors:
        return 0

    codes = tuple(
        code["code"]
        for error in errors
        if error
        for code in next(iter(error.values()))
    )

    messages = {
        rule["rule"]["name"]: rule["rule"]["message"]
        for rule in config.linter_rules
    }

    echo()
    width, _ = shutil.get_terminal_size()
    echo(
        f"{Fore.GREEN}{Style.BRIGHT}Statistics{Style.RESET_ALL}\n{Style.DIM}{'─' * width}{Style.RESET_ALL}"
    )

    if messages and codes:
        longest_code = len(max(messages.keys(), key=len))
        longest_count = len(
            str(max(Counter(codes).values(), key=_count_digits))
        )

        for code in sorted(Counter(codes).items()):
            code_space = (longest_code - len(code[0])) * " "
            count_space = (longest_count - _count_digits(code[1])) * " "

            echo(
                f"{Fore.YELLOW}{code[0]}{Fore.BLUE} {code_space}{code[1]}{Style.RESET_ALL} {count_space}{messages[code[0]]}"
            )

    return sum(Counter(codes).values())



"""djLint add indentation to html."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

import json5 as json
import regex as re

from ..helpers import (
    inside_ignored_block,
    is_ignored_block_closing,
    is_ignored_block_opening,
    is_safe_closing_tag,
    is_script_style_block_closing,
    is_script_style_block_opening,
)
from .attributes import format_attributes

if TYPE_CHECKING:
    from ..settings import Config


def indent_html(rawcode: str, config: Config) -> str:
    """Indent raw code."""
    if config.profile not in {"handlebars", "golang"}:
        # we can try to fix template tags. ignore handlebars
        # this should be done before indenting to line length
        # calc is preserved.

        def fix_tag_spacing(html: str, match: re.Match[str]) -> str:
            if inside_ignored_block(config, html, match):
                return match.group()

            return f"{match.group(1)} {match.group(2)} {match.group(3)}"

        """
        We should have tags like this:
        {{ tag }}
        {%- tag atrib -%}
        """
        func = partial(fix_tag_spacing, rawcode)

        rawcode = re.sub(
            r"({%-?\+?)[ ]*?(\w(?:(?!%}).)*?)[ ]*?(\+?-?%})", func, rawcode
        )

        rawcode = re.sub(
            r"({{)[ ]*?(\w(?:(?!}}).)*?)[ ]*?(\+?-?}})", func, rawcode
        )

    elif config.profile == "handlebars":

        def fix_handlebars_template_tags(
            html: str, match: re.Match[str]
        ) -> str:
            if inside_ignored_block(config, html, match):
                return match.group()

            return f"{match.group(1)} {match.group(2)}"

        func = partial(fix_handlebars_template_tags, rawcode)
        # handlebars templates
        rawcode = re.sub(r"({{#(?:each|if).+?[^ ])(}})", func, rawcode)

    rawcode_flat_list = re.split("\n", rawcode)

    indent = config.indent

    beautified_code = ""
    indent_level = 0
    in_set_tag = False
    is_raw_first_line = False
    in_script_style_tag = False
    is_block_raw = False
    jinja_replace_list = []

    slt_html = config.indent_html_tags

    # here using all tags cause we allow empty tags on one line
    always_self_closing_html = config.always_self_closing_html_tags

    # here using all tags cause we allow empty tags on one line
    slt_template = config.optional_single_line_template_tags

    # nested ignored blocks..
    ignored_level = 0

    for item in rawcode_flat_list:
        # if a raw tag first line
        if not is_block_raw and is_ignored_block_opening(config, item):
            is_raw_first_line = True

        # if a raw tag then start ignoring
        if is_ignored_block_opening(config, item):
            is_block_raw = True
            ignored_level += 1

        if is_script_style_block_opening(config, item):
            in_script_style_tag = True

        if is_safe_closing_tag(config, item):
            ignored_level -= 1
            ignored_level = max(ignored_level, 0)
            if is_block_raw and ignored_level == 0:
                is_block_raw = False

        if (
            re.findall(
                rf"^\s*?(?:{config.ignored_inline_blocks})",
                item,
                flags=re.IGNORECASE | re.VERBOSE | re.MULTILINE,
            )
            and not is_block_raw
        ) or (
            (
                re.findall(
                    rf"""^(?:[^<\s].*?)? # start of a line, optionally with some text
                    (?:
                        (?:<({slt_html})>)(?:.*?)(?:</(?:\1)>) # <span>stuff</span> >>>> match 1
                       |(?:<({slt_html})\b[^>]+?>)(?:.*?)(?:</(?:\2)>) # <span stuff>stuff</span> >>> match 2
                       |(?:<(?:{always_self_closing_html})\b[^>]*?/?>) # <img stuff />
                       |(?:<(?:{slt_html})\b[^>]*?/>) # <img />
                       |(?:{{%[ ]*?({slt_template})[ ]+?.*?%}})(?:.*?)(?:{{%[ ]+?end(?:\3)[ ]+?.*?%}}) # >>> match 3
                       |{config.ignored_inline_blocks}
                    )[ \t]*?
                    (?:
                    .*? # anything
                    (?: # followed by another slt
                        (?:<({slt_html})>)(?:.*?)(?:</(?:\4)>) # <span>stuff</span> >>>> match 1
                       |(?:<({slt_html})\b[^>]+?>)(?:.*?)(?:</(?:\5)>) # <span stuff>stuff</span> >>> match 2
                       |(?:<(?:{always_self_closing_html})\b[^>]*?/?>) # <img stuff />
                       |(?:<(?:{slt_html})\b[^>]*?/>) # <img />
                       |(?:{{%[ ]*?({slt_template})[ ]+?.*?%}})(?:.*?)(?:{{%[ ]+?end(?:\6)[ ]+?.*?%}}) # >>> match 3
                       |{config.ignored_inline_blocks}
                    )[ \t]*?
                    )*? # optional of course
                    [^<]*?$ # with no other tags following until end of line
                """,
                    item,
                    flags=re.IGNORECASE | re.VERBOSE | re.MULTILINE,
                )
            )
            and not is_block_raw
        ):
            tmp = (indent * indent_level) + item + "\n"

        # closing set tag
        elif (
            not config.no_set_formatting
            and re.search(
                r"^(?!.*\{\%).*%\}.*$",
                item,
                flags=re.IGNORECASE | re.MULTILINE | re.VERBOSE,
            )
            and not is_block_raw
            and in_set_tag
        ):
            indent_level = max(indent_level - 1, 0)
            in_set_tag = False
            tmp = (indent * indent_level) + item + "\n"

        # closing curly brace inside a set tag
        elif (
            not config.no_set_formatting
            and re.search(
                r"^[ ]*}|^[ ]*]",
                item,
                flags=re.IGNORECASE | re.MULTILINE | re.VERBOSE,
            )
            and not is_block_raw
            and in_set_tag
        ):
            indent_level = max(indent_level - 1, 0)
            tmp = (indent * indent_level) + item + "\n"

        # if unindent, move left
        elif (
            re.search(
                config.tag_unindent,
                item,
                flags=re.IGNORECASE | re.MULTILINE | re.VERBOSE,
            )
            and not is_block_raw
            and not is_safe_closing_tag(config, item)
            # and not ending in a slt like <span><strong></strong>.
            and not re.findall(
                rf"(<({slt_html})>)(.*?)(</(\2)>[^<]*?$)",
                item,
                flags=re.IGNORECASE | re.VERBOSE | re.MULTILINE,
            )
            and not re.findall(
                rf"(<({slt_html})\\b[^>]+?>)(.*?)(</(\2)>[^<]*?$)",
                item,
                flags=re.IGNORECASE | re.VERBOSE | re.MULTILINE,
            )
        ):
            # block to catch inline block followed by a non-break tag
            if re.findall(
                rf"(^<({slt_html})>)(.*?)(</(\2)>)",
                item,
                flags=re.IGNORECASE | re.VERBOSE | re.MULTILINE,
            ) or re.findall(
                rf"(^<({slt_html})\b[^>]+?>)(.*?)(</(\2)>)",
                item,
                flags=re.IGNORECASE | re.VERBOSE | re.MULTILINE,
            ):
                # unindent after instead of before
                tmp = (indent * indent_level) + item + "\n"
                indent_level = max(indent_level - 1, 0)
            else:
                indent_level = max(indent_level - 1, 0)
                tmp = (indent * indent_level) + item + "\n"

        elif (
            re.search(
                r"^" + str(config.tag_unindent_line),
                item,
                flags=re.IGNORECASE | re.MULTILINE | re.VERBOSE,
            )
            and not is_block_raw
        ):
            tmp = (indent * (indent_level - 1)) + item + "\n"

        # if indent, move right

        # opening set tag
        elif (
            not config.no_set_formatting
            and re.search(
                r"^([ ]*{%[ ]*?set)(?!.*%}).*$",
                item,
                flags=re.IGNORECASE | re.MULTILINE | re.VERBOSE,
            )
            and not is_block_raw
            and not in_set_tag
        ):
            tmp = (indent * indent_level) + item + "\n"
            indent_level += 1
            in_set_tag = True

        # opening curly brace inside a set tag
        elif (
            not config.no_set_formatting
            and re.search(
                r"(\{(?![^{}]*%[}\s])(?=[^{}]*$)|\[(?=[^\]]*$))",
                item,
                flags=re.IGNORECASE | re.MULTILINE | re.VERBOSE,
            )
            and not is_block_raw
            and in_set_tag
        ) or (
            re.search(
                r"^(?:" + str(config.tag_indent) + r")",
                item,
                flags=re.IGNORECASE | re.MULTILINE | re.VERBOSE,
            )
            and not is_block_raw
        ):
            tmp = (indent * indent_level) + item + "\n"
            indent_level += 1

        elif is_raw_first_line or (
            is_safe_closing_tag(config, item) and not is_block_raw
        ):
            tmp = (indent * indent_level) + item + "\n"

        elif is_block_raw or not item.strip():
            tmp = item + "\n"

        # otherwise, just leave same level
        elif not config.preserve_leading_space:
            # if we are not trying to preserve indenting
            # on text, the add it now.
            tmp = (indent * indent_level) + item + "\n"
        else:
            tmp = item + "\n"

        # if a opening raw tag then start ignoring.. only if there is no closing tag
        # on the same line
        if is_ignored_block_opening(config, item):
            is_block_raw = True
            is_raw_first_line = False

        # if a normal tag, we can try to expand attributes
        elif not is_block_raw:
            # get leading space, and attributes

            func = partial(format_attributes, config, item)

            tmp = re.sub(
                rf"(\s*?)(<(?:{config.indent_html_tags}))\s((?:\"[^\"]*\"|'[^']*'|{{[^}}]*}}|[^'\">{{}}\/])+?)(\s?/?>)",
                func,
                tmp,
                flags=re.VERBOSE | re.IGNORECASE,
            )

        # turn off raw block if we hit end - for one line raw blocks, but not an inline raw
        if is_ignored_block_closing(config, item) and (
            not in_script_style_tag
            or is_script_style_block_closing(config, item)
        ):
            in_script_style_tag = False
            if not is_safe_closing_tag(config, item):
                ignored_level -= 1
                ignored_level = max(ignored_level, 0)
            if ignored_level == 0:
                is_block_raw = False

        # detect the outer quotes for jinja
        if config.profile == "jinja":
            matches = re.findall(
                r"=([\"'])(\{\{[\s\S]*?\}\})\1", tmp, flags=re.MULTILINE
            )

            for match in matches:
                outer_quotes = match[0]
                inner_content = match[1]
                jinja_replace_list.append({
                    "outer_quote": outer_quotes,
                    "content": inner_content,
                })

        beautified_code += tmp

    # try to fix internal formatting of set tag
    def format_data(
        config: Config, contents: str, tag_size: int, leading_space: str
    ) -> str:
        try:
            # try to format the contents as json
            data = json.loads(contents)
            contents = json.dumps(
                data, trailing_commas=False, ensure_ascii=False, quote_keys=True
            )

            if tag_size + len(contents) >= config.max_line_length:
                # if the line is too long we can indent the json
                contents = json.dumps(
                    data,
                    indent=config.indent_size,
                    trailing_commas=False,
                    ensure_ascii=False,
                    quote_keys=True,
                )

        except Exception:
            # was not json.. try to eval as set
            try:
                # if contents is a python keyword, do not evaluate it.
                evaluated = (
                    str(eval(contents))  # noqa: S307
                    if contents != "object"
                    else contents
                )
                # need to unwrap the eval
                contents = (
                    evaluated[1:-1]
                    if contents[:1] != "(" and evaluated[:1] == "("
                    else evaluated
                )
            except Exception:
                contents = contents.strip()

        return (f"\n{leading_space}").join(contents.splitlines())

    def format_set(config: Config, html: str, match: re.Match[str]) -> str:
        if inside_ignored_block(config, html, match):
            return match.group()

        leading_space = match.group(1)
        open_bracket = match.group(2)
        tag = match.group(3)
        close_bracket = match.group(5)
        contents = match.group(4).strip()
        contents_split = contents.split("=", 1)

        if len(contents_split) > 1:
            contents = (
                contents_split[0].strip()
                + " = "
                + format_data(
                    config,
                    contents_split[-1],
                    len(f"{open_bracket} {tag}  {close_bracket}"),
                    leading_space,
                )
            )

        return f"{leading_space}{open_bracket} {tag} {contents} {close_bracket}"

    def format_function(config: Config, html: str, match: re.Match[str]) -> str:
        if inside_ignored_block(config, html, match):
            return match.group()

        leading_space = match.group(1)
        open_bracket = match.group(2)
        tag = match.group(3).strip()
        index = (match.group(5) or "").strip()
        close_bracket = match.group(6)
        contents = format_data(
            config,
            match.group(4).strip()[1:-1],
            len(f"{open_bracket} {tag}() {close_bracket}"),
            leading_space,
        )

        # # define cleaned match with both quote styles
        cleaned_match = f"{leading_space}{open_bracket} {tag}({contents}){index} {close_bracket}"

        if config.profile == "jinja":
            outer_quotes = None
            inner_quotes = None

            # Determine user quote type
            for jinja_content in jinja_replace_list:
                content = cleaned_match.replace(
                    '"', "'"
                )  # Replace double quotes
                if content == jinja_content.get("content"):
                    outer_quotes = jinja_content.get("outer_quote")
                    inner_quotes = "'" if outer_quotes == '"' else '"'
                    break
                content = cleaned_match.replace(
                    "'", '"'
                )  # Replace single quotes
                if content == jinja_content.get("content"):
                    outer_quotes = jinja_content.get("outer_quote")
                    inner_quotes = '"' if outer_quotes == "'" else "'"
                    break

            if outer_quotes is not None and inner_quotes is not None:
                # Replace all content inner quotes and remove trailing/leading spaces
                cleaned_contents = re.sub(
                    rf"(?<=\{re.escape(outer_quotes)})\s+|\s+(?=\{re.escape(outer_quotes)})",
                    "",
                    contents.replace(outer_quotes, inner_quotes),
                )

                # Update cleaned match
                cleaned_match = f"{leading_space}{open_bracket} {tag}({cleaned_contents}){index} {close_bracket}"
                cleaned_match = cleaned_match.strip()

        return cleaned_match

    if not config.no_set_formatting:
        func = partial(format_set, config, beautified_code)
        # format set contents
        beautified_code = re.sub(
            r"([ ]*)({%-?)[ ]*(set)[ ]+?((?:(?!%}).)*?)(-?%})",
            func,
            beautified_code,
            flags=re.IGNORECASE | re.MULTILINE | re.VERBOSE | re.DOTALL,
        )

    if not config.no_function_formatting:
        func = partial(format_function, config, beautified_code)
        # format function contents
        beautified_code = re.sub(
            r"([ ]*)({{-?\+?)[ ]*?((?:(?!}}).)*?\w)(\((?:\"[^\"]*\"|'[^']*'|[^\)])*?\)[ ]*)((?:\[[^\]]*?\]|\.[^\s]+)[ ]*)?((?:(?!}}).)*?-?\+?}})",
            func,
            beautified_code,
            flags=re.IGNORECASE | re.MULTILINE | re.VERBOSE | re.DOTALL,
        )

    if not config.preserve_blank_lines:
        beautified_code = beautified_code.lstrip()

    return beautified_code.rstrip() + "\n"



"""Build src file list."""

from __future__ import annotations

from typing import TYPE_CHECKING

import regex as re
from click import echo
from colorama import Fore

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from .settings import Config


def get_src(src: Iterable[Path], config: Config) -> list[Path]:
    """Get source files."""
    paths = []
    for item in src:
        # normalize path

        normalized_item = item.resolve()

        if (
            normalized_item.is_file()
            and no_pragma(config, normalized_item)
            and (
                (
                    config.use_gitignore
                    and not config.gitignore.match_file(normalized_item)
                )
                or not config.use_gitignore
            )
        ):
            paths.append(normalized_item)
            continue

        # remove leading . from extension
        extension = str(config.extension)
        extension = extension[1:] if extension.startswith(".") else extension

        paths.extend(
            x
            for x in normalized_item.glob(f"**/*.{extension}")
            if not re.search(config.exclude, x.as_posix(), flags=re.VERBOSE)
            and no_pragma(config, x)
            and (
                (config.use_gitignore and not config.gitignore.match_file(x))
                or not config.use_gitignore
            )
        )

    if not paths:
        echo(Fore.BLUE + "No files to check! 😢")

    return paths


html_patterns = (r"<!--\s*djlint\:on\s*-->",)
django_jinja_patterns = (
    r"\{#\s*djlint\:on\s*#\}",
    r"\{%\s*comment\s*%\}\s*djlint\:on\s*\{%\s*endcomment\s*%\}",
)
nunjucks_patterns = (r"\{#\s*djlint\:on\s*#\}",)
handlebars_patterns = (r"\{\{!--\s*djlint\:on\s*--\}\}",)
golang_patterns = (r"\{\{-?\s*/\*\s*djlint\:on\s*\*/\s*-?\}\}",)


def no_pragma(config: Config, this_file: Path) -> bool:
    """Verify there is no pragma present."""
    if not config.require_pragma:
        return True

    with this_file.open(encoding="utf-8") as open_file:
        first_line = open_file.readline()

        pragma_patterns = {
            "html": html_patterns,
            "django": django_jinja_patterns + html_patterns,
            "jinja": django_jinja_patterns + html_patterns,
            "nunjucks": nunjucks_patterns + html_patterns,
            "handlebars": handlebars_patterns + html_patterns,
            "golang": golang_patterns + html_patterns,
            "angular": html_patterns,
            "all": django_jinja_patterns
            + nunjucks_patterns
            + handlebars_patterns
            + golang_patterns
            + html_patterns,
        }

        return any(
            re.match(pattern, first_line)
            for pattern in pragma_patterns[config.profile]
        )



"""Condense HTML.

1. Put short html tags back on one line
2. Put short template tags back on one line
"""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

import regex as re

from ..helpers import (
    inside_ignored_block,
    inside_protected_trans_block,
    is_safe_closing_tag,
)

if TYPE_CHECKING:
    from ..settings import Config


def clean_whitespace(html: str, config: Config) -> str:
    """Compress back tags that do not need to be expanded."""
    # put empty tags on one line

    def strip_space(config: Config, html: str, match: re.Match[str]) -> str:
        """Trim leading whitespace."""
        # either inside a block, or this is a newline + closing block tag.
        # if it is a newline + closing block we can format it.

        if inside_ignored_block(
            config, html, match
        ) and not is_safe_closing_tag(config, match.group()):
            return match.group()

        # trimmed blocks should not be here.
        # we need to full html to check what type of
        # opening block it was - trimmed or not trimmed
        if inside_protected_trans_block(config, html[: match.end()], match):
            return match.group().rstrip()

        lines = len(re.findall(r"\n", match.group(2)))
        blank_lines = "\n" * lines
        if lines > config.max_blank_lines:
            blank_lines = "\n" * max(config.max_blank_lines, 0)
        return match.group(1) + blank_lines

    func = partial(strip_space, config, html)

    line_contents = r"(.*?)"
    trailing_contents = r"\n \t"

    if config.preserve_blank_lines:
        line_contents = r"([^\n]*?)"
        trailing_contents = r" \t"

    if not config.preserve_leading_space:
        # remove any leading/trailing space
        html = re.sub(
            rf"^[ \t]*{line_contents}([{trailing_contents}]*)$",
            func,
            html,
            flags=re.M,
        )

    else:
        # only remove leading space in front of tags
        # <, {%
        html = re.sub(
            rf"^[ \t]*((?:<|{{%).*?)([{trailing_contents}]*)$",
            func,
            html,
            flags=re.M,
        )
        html = re.sub(
            rf"^{line_contents}([{trailing_contents}]*)$",
            func,
            html,
            flags=re.M,
        )

    def add_blank_line_after(
        config: Config, html: str, match: re.Match[str]
    ) -> str:
        """Add break after if not in ignored block."""
        if inside_ignored_block(config, html, match):
            return match.group()

        # check that next line is not blank.
        if html[match.end() : match.end() + 1] != "\n":
            return match.group() + "\n"

        return match.group()

    func = partial(add_blank_line_after, config, html)

    # should we add blank lines after load tags?
    if config.blank_line_after_tag:
        for tag in config.blank_line_after_tag.split(","):
            html = re.sub(
                rf"((?:{{%\s*?{tag.strip()}\b[^}}]+?%}}\n?)+)",
                func,
                html,
                flags=re.IGNORECASE | re.MULTILINE | re.DOTALL,
            )

    def add_blank_line_before(
        config: Config, html: str, match: re.Match[str]
    ) -> str:
        """Add break before if not in ignored block and not first line in file."""
        if inside_ignored_block(config, html, match) or match.start() == 0:
            return match.group()

        return "\n" + match.group()

    func = partial(add_blank_line_before, config, html)

    # should we add blank lines before load tags?
    if config.blank_line_before_tag:
        for tag in config.blank_line_before_tag.split(","):
            html = re.sub(
                rf"(?<!^\n)((?:{{%\s*?{tag.strip()}\b[^}}]+?%}}\n?)+)",
                func,
                html,
                flags=re.IGNORECASE | re.MULTILINE | re.DOTALL,
            )

    # add line after yaml front matter

    def yaml_add_blank_line_after(html: str, match: re.Match[str]) -> str:
        """Add break after if not in ignored block."""
        if match.start() == 0 and not html.startswith("\n\n", match.end()):
            # verify there are not already blank lines
            return match.group() + "\n"

        return match.group()

    if not config.no_line_after_yaml:
        func = partial(yaml_add_blank_line_after, html)
        html = re.sub(
            r"(^---.+?---)$", func, html, flags=re.MULTILINE | re.DOTALL
        )

    return html


def condense_html(html: str, config: Config) -> str:
    """Put short tags back on a single line."""
    if config.preserve_leading_space:
        # if a user is attempting to reuse any leading
        # space for other purposes, we should not try to remove it.
        return html

    def condense_line(config: Config, html: str, match: re.Match[str]) -> str:
        """Put contents on a single line if below max line length."""
        if config.line_break_after_multiline_tag:
            # always force a break by pretending the line is too long.
            combined_length = config.max_line_length + 1
        else:
            combined_length = len(
                match.group(1).splitlines()[-1]
                + match.group(3)
                + match.group(4)
            )

        if (
            not inside_ignored_block(config, html, match)
            and combined_length < config.max_line_length
            and if_blank_line_after_match(config, match.group(3))
            and if_blank_line_before_match(config, match.group(3))
        ):
            return match.group(1) + match.group(3) + match.group(4)

        return match.group()

    def if_blank_line_after_match(config: Config, html: str) -> bool:
        """Check if there should be a blank line after."""
        if config.blank_line_after_tag:
            return not any(
                re.findall(
                    rf"((?:{{%\s*?{tag}[^}}]+?%}}\n?)+)",
                    html,
                    flags=re.IGNORECASE | re.MULTILINE | re.DOTALL,
                )
                for tag in (
                    x.strip() for x in config.blank_line_after_tag.split(",")
                )
            )
        return True

    def if_blank_line_before_match(config: Config, html: str) -> bool:
        """Check if there should be a blank line before."""
        if config.blank_line_before_tag:
            return not any(
                re.findall(
                    rf"((?:{{%\s*?{tag}[^}}]+?%}}\n?)+)",
                    html,
                    flags=re.IGNORECASE | re.MULTILINE | re.DOTALL,
                )
                for tag in (
                    x.strip() for x in config.blank_line_before_tag.split(",")
                )
            )
        return True

    # add blank lines before tags
    func = partial(condense_line, config, html)

    # put short single line tags on one line
    html = re.sub(
        rf"(<({config.optional_single_line_html_tags})\b(?:\"[^\"]*\"|'[^']*'|{{[^}}]*}}|[^'\">{{}}])*>)\s*([^<\n]*?)\s*?(</(\2)>)",
        func,
        html,
        flags=re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE,
    )

    # put short template tags back on one line. must have leading space
    # jinja +%} and {%+ intentionally omitted.
    return re.sub(
        rf"((?:\s|^){{%-?[ ]*?({config.optional_single_line_template_tags})\b(?:(?!\n|%}}).)*?%}})\s*([^%\n]*?)\s*?({{%-?[ ]+?end(\2)[ ]*?%}})",
        func,
        html,
        flags=re.IGNORECASE | re.MULTILINE | re.VERBOSE,
    )



"""rule H025: Check for orphans html tags."""

from __future__ import annotations

import copy
from itertools import chain
from typing import TYPE_CHECKING

import regex as re

from ..helpers import (
    inside_ignored_linter_block,
    inside_ignored_rule,
    overlaps_ignored_block,
)
from ..lint import get_line

if TYPE_CHECKING:
    from typing_extensions import Any

    from ..lint import LintError
    from ..settings import Config


def run(
    rule: dict[str, Any],
    config: Config,
    html: str,
    filepath: str,
    line_ends: list[dict[str, int]],
    *args: Any,
    **kwargs: Any,
) -> tuple[LintError, ...]:
    """Check for orphans html tags."""
    open_tags: list[re.Match[str]] = []
    orphan_tags: list[re.Match[str]] = []

    for match in re.finditer(
        r"<(/?(\w+))\s*(" + config.attribute_pattern + r"|\s*)*\s*?>",
        html,
        flags=re.VERBOSE,
    ):
        if match.group(1) and not re.search(
            rf"^/?{config.always_self_closing_html_tags}\b",
            match.group(1),
            flags=re.I | re.X,
        ):
            # close tags should equal open tags
            if match.group(1)[0] != "/":
                open_tags.insert(0, match)
            else:
                for i, tag in enumerate(copy.deepcopy(open_tags)):
                    if tag.group(2) == match.group(1)[1:]:
                        open_tags.pop(i)
                        break
                else:
                    # there was no open tag matching the close tag
                    orphan_tags.append(match)

    return tuple(
        {
            "code": rule["name"],
            "line": get_line(match.start(), line_ends),
            "match": match.group().strip()[:20],
            "message": rule["message"],
        }
        for match in chain(open_tags, orphan_tags)
        if (
            not overlaps_ignored_block(config, html, match)
            and not inside_ignored_rule(config, html, match, rule["name"])
            and not inside_ignored_linter_block(config, html, match)
        )
    )



"""Expand html.

1. put html tags on individual lines, if needed.
2. put template tags on individual lines, if needed.
"""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

import regex as re

from ..helpers import inside_ignored_block, inside_template_block

if TYPE_CHECKING:
    from ..settings import Config


def expand_html(html: str, config: Config) -> str:
    """Split single line html into many lines based on tags."""

    def add_html_line(out_format: str, match: re.Match[str]) -> str:
        """Add whitespace.

        Do not add whitespace if the tag is in a non indent block.

        Do not add whitespace if the tag is a in a template block
        """
        if inside_ignored_block(config, html, match):
            return match.group(1)

        if inside_template_block(config, html, match):
            return match.group(1)

        if out_format == "\n%s" and match.start() == 0:
            return match.group(1)

        return out_format % match.group(1)

    html_tags = config.break_html_tags

    add_left = partial(add_html_line, "\n%s")
    add_right = partial(add_html_line, "%s\n")

    break_char = config.break_before

    # html tags - break before
    html = re.sub(
        rf"{break_char}\K(</?(?:{html_tags})\b(\"[^\"]*\"|'[^']*'|{{[^}}]*}}|[^'\">{{}}])*>)",
        add_left,
        html,
        flags=re.IGNORECASE | re.VERBOSE,
    )

    # html tags - break after
    html = re.sub(
        rf"(</?(?:{html_tags})\b(\"[^\"]*\"|'[^']*'|{{[^}}]*}}|[^'\">{{}}])*>)(?!\s*?\n)(?=[^\n])",
        add_right,
        html,
        flags=re.IGNORECASE | re.VERBOSE,
    )

    # template tag breaks
    def should_i_move_template_tag(
        out_format: str, match: re.Match[str]
    ) -> str:
        # ensure template tag is not inside an html tag and also not the first line of the file
        if inside_ignored_block(config, html, match):
            return match.group(1)

        if not re.findall(
            r"\<(?:"
            + str(config.indent_html_tags)
            # added > as not allowed inside a "" or '' to prevent invalid wild html matches
            # for issue #640
            + r")\b(?:\"[^\">]*\"|'[^'>]*'|{{[^}]*}}|{%[^%]*%}|{\#[^\#]*\#}|[^>{}])*?"
            + re.escape(match.group(1))
            + "$",
            html[: match.end()],
            flags=re.MULTILINE | re.VERBOSE,
        ):
            if out_format == "\n%s" and match.start() == 0:
                return match.group(1)
            return out_format % match.group(1)

        return match.group(1)

    # template tags
    # break before
    html = re.sub(
        break_char
        + r"\K((?:{%|{{\#)[ ]*?(?:"
        + config.break_template_tags
        + ")[^}]+?[%}]})",
        partial(should_i_move_template_tag, "\n%s"),
        html,
        flags=re.IGNORECASE | re.MULTILINE | re.VERBOSE,
    )

    # break after
    return re.sub(
        r"((?:{%|{{\#)[ ]*?(?:"
        + config.break_template_tags
        + ")[^}]+?[%}]})(?=[^\n])",
        partial(should_i_move_template_tag, "%s\n"),
        html,
        flags=re.IGNORECASE | re.MULTILINE | re.VERBOSE,
    )



"""Djlint html linter."""

from __future__ import annotations

import importlib
from collections.abc import Sequence
from typing import TYPE_CHECKING

import regex as re

from .helpers import (
    inside_ignored_linter_block,
    inside_ignored_rule,
    overlaps_ignored_block,
)

if TYPE_CHECKING:
    from collections.abc import Mapping
    from pathlib import Path

    from typing_extensions import TypedDict

    from .settings import Config

    class LintError(TypedDict):
        code: str
        line: str
        match: str
        message: str


flags = {
    "re.A": re.A,
    "re.ASCII": re.ASCII,
    "re.I": re.I,
    "re.IGNORECASE": re.IGNORECASE,
    "re.M": re.M,
    "re.MULTILINE": re.MULTILINE,
    "re.S": re.S,
    "re.DOTALL": re.DOTALL,
    "re.X": re.X,
    "re.VERBOSE": re.VERBOSE,
    "re.L": re.L,
    "re.LOCALE": re.LOCALE,
}


def build_flags(flag_list: str | int) -> int:
    """Build list of regex flags."""
    if isinstance(flag_list, int):
        return flag_list
    split_flags = flag_list.split("|")
    combined_flags = 0
    for flag in split_flags:
        combined_flags |= flags[flag.strip()]
    return combined_flags


def get_line(start: int, line_ends: Sequence[Mapping[str, int]]) -> str:
    """Get the line number and index of match."""
    line = next(pair for pair in line_ends if pair["end"] > start)

    return "{}:{}".format(line_ends.index(line) + 1, start - line["start"])


def linter(
    config: Config, html: str, filename: str, filepath: str
) -> dict[str, list[LintError]]:
    """Lint a html string."""
    errors: dict[str, list[LintError]] = {filename: []}
    # build list of line ends for file
    line_ends = [
        {"start": m.start(), "end": m.end()}
        for m in re.finditer(r"(?:.*\n)|(?:[^\n]+$)", html)
    ]

    ignored_rules: set[str] = set()

    # remove ignored rules for file
    for pattern, rules in config.per_file_ignores.items():
        if re.search(pattern, filepath, flags=re.VERBOSE):
            ignored_rules.update(x.strip() for x in rules.split(","))

    for rule in config.linter_rules:
        rule = rule["rule"]  # noqa: PLW2901

        # skip ignored rules
        if rule["name"] in ignored_rules:
            continue

        # rule based on python module
        if "python_module" in rule:
            rule_module = importlib.import_module(rule["python_module"])
            module_errors = rule_module.run(
                rule=rule,
                config=config,
                html=html,
                filepath=filepath,
                line_ends=line_ends,
            )
            if not isinstance(module_errors, Sequence):
                msg = (
                    f"Error: {rule['name']} python_module run() should return"
                    " a sequence of dict with keys: code, line, match, message."
                )
                raise AssertionError(msg)
            errors[filename].extend(module_errors)

        # rule based on patterns
        else:
            for pattern in rule["patterns"]:
                for match in re.finditer(
                    pattern,
                    html,
                    flags=build_flags(rule.get("flags", "re.DOTALL")),
                ):
                    if (
                        not overlaps_ignored_block(config, html, match)
                        and not inside_ignored_rule(
                            config, html, match, rule["name"]
                        )
                        and not inside_ignored_linter_block(config, html, match)
                    ):
                        errors[filename].append({
                            "code": rule["name"],
                            "line": get_line(match.start(), line_ends),
                            "match": match.group().strip()[:20],
                            "message": rule["message"],
                        })

    # remove duplicate matches
    for filename, error_dict in errors.items():  # noqa: PLR1704
        unique_errors = []
        for dict_ in error_dict:
            if dict_ not in unique_errors:
                unique_errors.append(dict_)
        errors[filename] = unique_errors
    return errors


def lint_file(config: Config, this_file: Path) -> dict[str, list[LintError]]:
    """Check file for formatting errors."""
    filename = str(this_file)

    html = this_file.read_text(encoding="utf-8")

    return linter(config, html, filename, this_file.as_posix())
